// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "PrefabTool.hpp"

using namespace Dawn;

std::vector<std::string> PrefabTool::getRequiredFlags() {
  return { "input", "output", "sources" };
}

std::map<std::string, std::string> PrefabTool::getOptionalFlags() {
  return std::map<std::string, std::string>();
}

int32_t PrefabTool::start() {
  File input = File(flags["input"]);
  if(!input.exists()) {
    std::cout << "Input file does not exist!" << std::endl;
    return 1;
  }

  std::string data;
  if(!input.readString(&data)) {
    std::cout << "Failed to read input file!" << std::endl;
    return 1;
  }

  struct PrefabRegistry registry;
  registry.sources = File::normalizeSlashes(flags["sources"]);

  auto xml = Xml::load(data);
  std::string error;
  struct Prefab prefab;
  prefab.registry = &registry;
  auto result = ((PrefabParser()).parse(&xml, &prefab, &error));
  if(result != 0) {
    std::cout << "Failed to parse prefab: " << error << std::endl;
    return result;
  }

  // Generate output
  std::vector<std::string> outputData;
  PrefabGen::generate(&outputData, &prefab, "");

  // Load output file from name and type
  File outputFile = File(flags["output"] + "/prefabs/" + prefab.type + "/" + prefab.name + ".hpp");
  if(!outputFile.mkdirp()) {
    std::cout << "Failed to create output directory!" << std::endl;
    return 1;
  }

  // Combine vector into single string
  std::string outputDataStr = "";
  auto it = outputData.begin();
  while(it != outputData.end()) {
    outputDataStr += *it + "\n";
    ++it; 
  }
  
  if(!outputFile.writeString(outputDataStr)) {
    std::cout << "Failed to write output file!" << std::endl;
    return 1;
  }

  return 0;
}

//

void PrefabGen::generate(
  std::vector<std::string> *out,
  struct Prefab *info,
  std::string tabs
) {
  struct ClassGenInfo classInfo;
  classInfo.clazz = info->name;
  classInfo.extend = "SceneItemPrefab<" + info->name + ">";
  classInfo.constructorArgs = "Scene *scene, sceneitemid_t id";
  classInfo.extendArgs = "scene, id";

  struct MethodGenInfo methodAssets;
  methodAssets.name = "prefabAssets";
  methodAssets.isStatic = true;
  methodAssets.type = "std::vector<Asset*>";
  methodAssets.args = "AssetManager *man";
  line(&methodAssets.body, "std::vector<Asset*> assets;", "");

  struct MethodGenInfo methodInit;
  methodInit.name = "prefabInit";
  methodInit.isOverride = true;
  methodInit.args = "AssetManager *man";

  classInfo.includes.push_back("prefab/SceneItemPrefab.hpp");

  // Includes
  int32_t componentNumber = 0;
  auto itChildren = info->components.begin();
  while(itChildren != info->components.end()) {
    auto c = *itChildren;
    std::string name = "c" + std::to_string(componentNumber++);
    bool_t init = true;
    if(c.ref.size() > 0) {
      init = false;
      name = c.ref;
      line(&classInfo.publicProperties, c.type + " *" + c.ref + " = nullptr;", "");
    }

    // Initialize
    line(&methodInit.body, (init ? "auto " : "") + name + " = this->addComponent<" + c.type + ">();", "");

    // Now set each property
    auto itValues = c.values.begin();
    while(itValues != c.values.end()) {
      line(&methodInit.body, name + "->" + itValues->first + " = " + itValues->second + ";", "");
      ++itValues;
    }
    line(&methodInit.body, "", "");

    classInfo.includes.push_back(c.include);
    ++itChildren;
  }
  
  // Seal methods
  line(&methodAssets.body, "return assets;", "");

  // Add in methods
  CodeGen::methodGen(&classInfo.publicCode, methodAssets);
  line(&classInfo.publicCode, "", "");
  CodeGen::methodGen(&classInfo.publicCode, methodInit);

  CodeGen::classGen(out, classInfo);
}